@@ -3,17 +3,19 @@
 and sending of data to Devo """
 import errno
 import logging
+import select
 import socket
 import ssl
 import sys
 import time
 import zlib
-from _socket import SHUT_RDWR
+from _socket import SHUT_WR, SHUT_RD
 from enum import Enum
 from pathlib import Path
 
 import pem
 from OpenSSL import SSL, crypto
+from ssl import SSLWantReadError, SSLWantWriteError
 
 from devo.common import Configuration, get_log, get_stream_handler
 from .transformsyslog import (COMPOSE, COMPOSE_BYTES, FACILITY_USER, FORMAT_MY,
@@ -370,6 +372,9 @@ class Sender(logging.Handler):
         if isinstance(config, SenderConfigTCP):
             self.__connect_tcp_socket()
 
+    def __del__(self):
+        self.close()
+
     def __connect(self):
         if isinstance(self._sender_config, SenderConfigSSL):
             self.__connect_ssl()
@@ -392,6 +397,7 @@ class Sender(logging.Handler):
                 ERROR_MSGS.TCP_CONN_ESTABLISHMENT_SOCKET % str(error)) from error
 
         self.timestart = int(round(time.time() * 1000))
+        self.socket.setblocking(False)
 
     def __connect_ssl(self):
         """
@@ -453,6 +459,7 @@ class Sender(logging.Handler):
                                   % (repr(self.socket.getpeername()),
                                      str(self.reconnection)))
             self.timestart = int(round(time.time() * 1000))
+            self.socket.setblocking(False)
 
         except socket.error as error:
             self.close()
@@ -558,6 +565,12 @@ class Sender(logging.Handler):
             self.close()
             return False
 
+        # If no data, EOF and channel is closed
+        if self.__check_EOF():
+            # Restart connection
+            self.close()
+            return False
+
         return True
 
     def close(self):
@@ -566,11 +579,13 @@ class Sender(logging.Handler):
         """
         if self.socket is not None:
             try:
-                self.socket.shutdown(SHUT_RDWR)
+                self.socket.shutdown(SHUT_WR)
+                self.__wait_for_EOF()
             except Exception:  # Try else continue
                 logging.exception(ERROR_MSGS.CLOSING_ERROR)
-            self.socket.close()
-            self.socket = None
+            finally:
+                self.socket.close()
+                self.socket = None
 
     @staticmethod
     def __encode_multiline(record):
@@ -600,14 +615,126 @@ class Sender(logging.Handler):
         for iteration in range(0, total):
             part = record[int(iteration * 4096):
                           int((iteration + 1) * 4096)]
+            self.__sendall(part)
             self.last_message = int(time.time())
-            if self.socket.sendall(part) is not None:
-                raise DevoSenderException(ERROR_MSGS.SEND_ERROR)
             sent += len(part)
         if sent == 0:
             raise DevoSenderException(ERROR_MSGS.SEND_ERROR)
         return sent
 
+    def __sendall(self, content):
+        """
+        Send content to endpoint dealing with blocking socket and SSL wrapper
+        :param content: The content to be sent as event
+        :raises DevoSenderException: if data cannot be sent or timeout is reached before sending it
+        :return: No expected return
+        """
+        then = time.time()
+        # Is the channel ready for writing?
+        if select.select([], [self.socket], [], self.socket_timeout)[1]:
+            while True:
+                try:
+                    # Write it
+                    if self.socket.sendall(content) is not None:
+                        raise DevoSenderException(str(ERROR_MSGS.SEND_ERROR))
+                    else:
+                        return
+                except BlockingIOError as exc:
+                    # This is not blocking socket
+                    if exc.errno == errno.ECONNRESET:
+                        # A TCP RST implies error in channel
+                        raise IOError("Connection reset by endpoint") from exc
+                    elif exc.errno not in [errno.EAGAIN, errno.EWOULDBLOCK]:
+                        # Any other error but EAGAIN and EWOULDBLOCK here
+                        raise IOError("Error while accessing socket") from exc
+                    else:
+                        # errno.EAGAIN means nothing to get, but socket working
+                        pass
+                except SSLWantWriteError:
+                    # If the data is ready at socket OS level but not at SSL wrapper level, this exception may raise
+                    pass
+                # Try while timeout not reached
+                if time.time() - self.socket_timeout >= then:
+                    raise DevoSenderException(ERROR_MSGS.ERROR_AFTER_TIMEOUT)
+        else:
+            raise DevoSenderException(ERROR_MSGS.ERROR_AFTER_TIMEOUT)
+
+    def __check_EOF(self):
+        """
+        Checks for EOF of the downstream channell to check whether endpoint closed connection
+        If the channel was closed by the other endpoint (ingestion balancer)
+        the reading of the download channel will return EOF. This is
+        checked by reading a ready buffer but getting no data, empty bytes
+        :return: Whether the EOF is detected in downstream (True) or not (False)
+        """
+        # Is the channel ready for reading?
+        if select.select([self.socket], [], [], 0)[0]:
+            while True:
+                try:
+                    # Read it
+                    buf = self.socket.recv(1)
+                    # If no data, EOF and channel is closed
+                    return buf == b''
+                except BlockingIOError as exc:
+                    # This is not blocking socket
+                    if exc.errno == errno.ECONNRESET:
+                        # A TCP RST implies error in channel
+                        raise IOError("Connection reset by endpoint") from exc
+                    elif exc.errno not in [errno.EAGAIN, errno.EWOULDBLOCK]:
+                        # Any other error but EAGAIN and EWOULDBLOCK here
+                        raise IOError("Error while accessing socket") from exc
+                    else:
+                        # errno.EAGAIN means nothing to get, but socket working
+                        # Nothing to read, everything is ok
+                        return False
+                except SSLWantReadError:
+                    # If the data is ready at socket OS level but not at SSL wrapper level, this exception may raise
+                    # Nothing to read, everything is ok
+                    return False
+        else:
+            return False
+
+    def __wait_for_EOF(self):
+        """
+        Wait for the endpoint to close the downstream channel after client closed the upstream one
+        The downstream channel is closed by the other endpoint, ingestion balancer,
+        by sending EOF. This is checked by reading a ready buffer but getting no data, empty bytes
+        :raises DevoSenderException: if timeout is reached before sending it
+        :return: Array of bytes with all the data send by endpoint until closing
+        """
+        then = time.time()
+        # Wait for the channel to be ready for reading (with timeout)
+        if select.select([self.socket], [], [], self.socket_timeout)[0]:
+            bytes = bytearray()
+            while True:
+                try:
+                    # Read it
+                    buf = self.socket.recv(1)
+                    # If no data, EOF and channel is closed
+                    if buf == b'':
+                        return bytes
+                    else:
+                        bytes.extend(buf)
+                except BlockingIOError as exc:
+                    # This is not blocking socket
+                    if exc.errno == errno.ECONNRESET:
+                        # A TCP RST implies error in channel
+                        raise IOError("Connection reset by endpoint") from exc
+                    elif exc.errno not in [errno.EAGAIN, errno.EWOULDBLOCK]:
+                        # Any other error but EAGAIN and EWOULDBLOCK here
+                        raise IOError("Error while accessing socket") from exc
+                    else:
+                        # errno.EAGAIN means nothing to get, but socket working
+                        pass
+                except SSLWantReadError:
+                    # If the data is ready at socket OS level but not at SSL wrapper level, this exception may raise
+                    pass
+                # Try while timeout not reached
+                if time.time() - self.socket_timeout >= then:
+                    raise DevoSenderException(ERROR_MSGS.ERROR_AFTER_TIMEOUT)
+        else:
+            raise DevoSenderException(ERROR_MSGS.ERROR_AFTER_TIMEOUT)
+
     def send_raw(self, record, multiline=False, zip=False):
         """
         Send raw messages to the collector
@@ -625,9 +752,8 @@ class Sender(logging.Handler):
                     if not multiline and not zip:
                         msg = self.__encode_record(record)
                         sent = len(msg)
+                        self.__sendall(msg)
                         self.last_message = int(time.time())
-                        if self.socket.sendall(msg) is not None:
-                            raise DevoSenderException(ERROR_MSGS.SEND_ERROR)
                         return 1
                     if multiline:
                         record = self.__encode_multiline(record)
